#include <stdio.h>
#include <dos.h>
#include <mem.h>
#include <ctype.h>
#include <string.h>
#include "systable.h"

#define TRUE    1
#define FALSE   !TRUE

void find_sign (unsigned, unsigned, char *);

void bios_info (struct BIOS *bios)
{
    void             *signature;
    int              i;
    union    REGS    inregs, outregs;
    struct  SREGS    segregs;
    struct   SDT     *sdt;
    struct BIOS_INFO *bios_base;

    memcpy (bios->date, BIOS_DATE, 8);
    bios->date [8] = '\0';
    if ((!(bios->date [2] == '/')) | (!(bios->date [5] == '/')))
        strcpy (bios->date, "Unknown");

    find_sign (0xF000, 0x0000, bios->sign);
    if ((strncmp (bios->sign, "Signature not found", 19)) == 0)
        {
        find_sign (0xF000, 0x8000, bios->sign);
        }

    find_sign (0xC000, 0x8000, bios->hard_disk_sign);
    if ((strncmp (bios->hard_disk_sign, "Signature not found", 19)) == 0)
        {
        strcpy (bios->hard_disk_sign, "Hard disk supported by native ROM or not present");
        }

    bios->sysid_model = (byte)peekb(BIOS_SYSID_SEG, BIOS_SYSID_OFF);

    inregs.h.ah = 0xC0;                            /* Return System Config. */
    int86x (0x15, &inregs, &outregs, &segregs);        /* Issue interrupt */
    sdt = MK_FP(segregs.es, outregs.x.bx);        /* make it a far pointer */

    if (outregs.x.cflag != 0)
        {
        strcpy (bios->bus, "Industry Standard");
        }
    else
        {
        strcpy (bios->bus, (sdt->features.isa_mca ?
                    "Micro Channel" :
                    "Industry Standard"));
        }

    bios->sysid_submodel = sdt->submodel;
    bios->sysid_revision = sdt->rev_level;

    bios->AT = FALSE;

    strcpy (bios->chip2,    (sdt->features.chip_2   ? "Yes" : " No"));
    strcpy (bios->int09,    (sdt->features.kbd_int  ? "Yes" : " No"));
    strcpy (bios->ext_wait, (sdt->features.ext_wait ? "Yes" : " No"));

    switch (bios->sysid_model)
        {
        case 0xFF:  strcpy (bios->sysid_char, "PC");
                    break;
        case 0xFE:  if (strncmp (bios->date, "11/08/82", 8) == 0)
                        strcpy (bios->sysid_char, "XT or Portable");
                    else
                        strcpy (bios->sysid_char, "Compaq DeskPro");
                    break;
        case 0xFD:  strcpy (bios->sysid_char, "PCjr");
                    break;
        case 0xFC:  switch (bios->sysid_submodel)
                        {
                        case 0x00:
                        case 0x01:  bios->AT = TRUE;
                                    strcpy (bios->sysid_char, "AT");
                                    break;
                        case 0x02:  strcpy (bios->sysid_char, "PC/XT286");
                                    break;
                        case 0x04:  bios->AT = TRUE;
                                    strcpy (bios->sysid_char, "PS/2 Model 50");
                                    break;
                        case 0x05:  bios->AT = TRUE;
                                    strcpy (bios->sysid_char, "PS/2 Model 60");
                                    break;
                        }
                    break;
        case 0xFA:  strcpy (bios->sysid_char, "PS/2 Model 30");
                    break;
        case 0xF9:  strcpy (bios->sysid_char, "PC Convertible");
                    break;
        case 0xF8:  bios->AT = TRUE;
                    strcpy (bios->sysid_char, "PS/2 Model 80");
                    break;
        case 0xB6:  strcpy (bios->sysid_char, "HP110 Portable");
                    break;
        case 0x9A:  strcpy (bios->sysid_char, "Compaq Portable Plus");
                    break;
        case 0x4B:  strcpy (bios->sysid_char, "MegaBIOS ROM");
                    break;
        case 0x2D:  strcpy (bios->sysid_char, "Compaq Portable");
                    break;
        default:    strcpy (bios->sysid_char, "Unknown machine type");
                    break;
        }

    bios_base = MK_FP (BIOS_SEG, BIOS_OFF);

    if (bios_base->com1 != 0)
        sprintf (bios->com1, "%04X", bios_base->com1);
    else (strcpy (bios->com1, " N/A"));

    if (bios_base->com2 != 0)
        sprintf (bios->com2, "%04X", bios_base->com2);
    else (strcpy (bios->com2, " N/A"));

    if (bios_base->com3 != 0)
        sprintf (bios->com3, "%04X", bios_base->com3);
    else (strcpy (bios->com3, " N/A"));

    if (bios_base->com4 != 0)
        sprintf (bios->com4, "%04X", bios_base->com4);
    else (strcpy (bios->com4, " N/A"));

    if (bios_base->lpt1 != 0)
        sprintf (bios->lpt1, "%04X", bios_base->lpt1);
    else (strcpy (bios->lpt1, " N/A"));

    if (bios_base->lpt2 != 0)
        sprintf (bios->lpt2, "%04X", bios_base->lpt2);
    else (strcpy (bios->lpt2, " N/A"));

    if (bios_base->lpt3 != 0)
        sprintf (bios->lpt3, "%04X", bios_base->lpt3);
    else (strcpy (bios->lpt3, " N/A"));

    if (bios_base->lpt4 != 0)
        sprintf (bios->lpt4, "%04X", bios_base->lpt4);
    else (strcpy (bios->lpt4, " N/A"));

    bios->n_floppies = bios_base->hardware_1.n_floppies + 1;
    bios->n_fixed_disks = bios_base->n_fixed_disks;

    switch (bios_base->drive0_ms.data_rate)
        {
        case 0: bios->drive0_data_rate = 500;
                break;
        case 1: bios->drive0_data_rate = 300;
                break;
        case 2: bios->drive0_data_rate = 250;
                break;
        default: bios->drive0_data_rate = 0;
        }
    switch (bios_base->drive1_ms.data_rate)
        {
        case 0: bios->drive1_data_rate = 500;
                break;
        case 1: bios->drive1_data_rate = 300;
                break;
        case 2: bios->drive1_data_rate = 250;
                break;
        default: bios->drive1_data_rate = 0;
        }
}
