struct  MS
    {
    unsigned media_state         :3; /* not currently documented */
    unsigned reserved            :1;
    unsigned media_established   :1;
    unsigned double_step_required:1;
    unsigned data_rate           :2; /* 00 = 500KB/s, 01 = 300KB/s, 10 = 250KB/s */
    };

struct BIOS_INFO
    {
    word    com1;
    word    com2;
    word    com3;
    word    com4;
    word    lpt1;
    word    lpt2;
    word    lpt3;
    word    lpt4;
    struct
        {
        unsigned is_floppies :1;
        unsigned is_math     :1;
        unsigned is_mouse    :1;    /* PS/2 only */
        unsigned a           :1;    /* reserved  */
        unsigned video_mode  :2;    /* 01=CO40, 10=CO80, 11=BW80 */
        unsigned n_floppies  :2;
        }   hardware_1;
    struct
        {
        unsigned a           :1;    /* reserved */
        unsigned n_rs232     :3;
        unsigned is_joystick :1;
        unsigned is_modem    :1;    /* Convertible only */
        unsigned n_lpt       :2;
        }   hardware_2;
    byte    POST_status;            /* Convertible only */
    word    mem_size;
    word    bios_40_15;             /* reserved */
    struct
        {
        unsigned r_shift     :1;
        unsigned l_shift     :1;
        unsigned ctrl        :1;
        unsigned alt         :1;
        unsigned scroll_lock :1; /* scroll lock mode */
        unsigned num_lock    :1; /* num lock mode */
        unsigned caps_lock   :1; /* caps lock mode */
        unsigned insert      :1; /* insert mode */
        }   kbd_1;
    struct
        {
        unsigned l_ctrl      :1; /* Left ctrl key is down */
        unsigned l_alt       :1;
        unsigned SysReq      :1;
        unsigned pause       :1; /* pause mode active */
        unsigned scroll_lock :1; /* scroll lock key down */
        unsigned num_lock    :1; /* num lock key down */
        unsigned caps_lock   :1; /* caps lock key down */
        unsigned insert      :1; /* insert key down */
        }   kbd_2;
    byte    alt_keypad;     /* alternate keypad entry */
    word    kbd_head;       /* keyboard buffer head */
    word    kbd_tail;       /* keyboard buffer tail */
	byte    kbd_buff[32];   /* keyboard buffer */
    struct
        {
        unsigned drive0      :1;
        unsigned drive1      :1;
        unsigned drive2      :1;
        unsigned drive3      :1;
        unsigned reserved    :3;
        unsigned intrupt     :1;
        } recallibrate_status;
    struct
        {
        unsigned drive0      :1;
        unsigned drive1      :1;
        unsigned drive2      :1;
        unsigned drive3      :1;
        unsigned selected    :2;
        unsigned reserved    :1;
        unsigned write_read  :1;
        } motor_status;
    byte    motor_off_count;
    struct
        {
		unsigned bad_drive   :1; /* invalid drive parameter */
        unsigned bad_address :1; /* address mark not found */
        unsigned sector_nf   :1; /* sector not found */
        unsigned DMA_overrun :1; /* DMA overrun on operation */
        unsigned CRC_error   :1; /* CRC error on diskette read */
        unsigned general_err :1; /* General controller failure */
        unsigned seek_failed :1; /* Seek operation failed */
        unsigned not_ready   :1; /* Drive not ready */
        } floppy_prev_op_status;
	byte    floppy_controller_status[7];
    byte    display_mode;
    word    n_cols;
    word    regen_len;
    word    regen_addr;
    word    cursor_pos_page_1;
    word    cursor_pos_page_2;
    word    cursor_pos_page_3;
    word    cursor_pos_page_4;
    word    cursor_pos_page_5;
    word    cursor_pos_page_6;
    word    cursor_pos_page_7;
    word    cursor_pos_page_8;
    word    cursor_type;
    byte    current_page;
    word    video_controller_base_addr;
    byte    curr_3x8_register;
    byte    curr_3x9_register;
    byte far *reset_code;   /* PS/2 only (except Model 30) */
    byte    bios_40_6B;     /* reserved */
    dblwd   timer_count;    /* timer counter */
    byte    timer_oflow_flag;
    byte    break_key_state;
    word    reset_flag;     /* 0x1234 = bypass memory test
                             * 0x4321 = preserve memory (PS/2)
                             * 0x5678 = system suspended (convertible)
                             * 0x9ABC = manufacturer test (convertible)
                             * 0xABCD = system post loop (convertible)
                            */
    byte    prev_op_status; /*      H/F (Hard/Floppy/Both)
                             * 00 = x x No error
                             * 01 = x x Bad Command
                             * 02 = x x Address mark not found
                             * 03 =   x Write protect error
                             * 04 = x x Sector not found
                             * 05 = x   Reset failed
                             * 06 =   x Diskette was removed
                             * 07 = x   Bad parameter table
                             * 08 =   x DMA overrun
                             * 09 = x x Attempted to cross 64k boundary on DMA
                             * 0A = x   Bad sector flag
                             * 0B = x   Bad cylinder detected ~
                             * 0C =   x Requested Media type not found ~
                             * 0D = x   Invalid # sectors in format ~
                             * 0E = x   Control data address mark detected ~
                             * 0F = x   DMA arbitration level out of range ~
                             * 10 = x x CRC or ECC error on disk read
                             * 11 = x   ECC corrected data error
                             * 20 = x x Controller failed
                             * 40 = x x Seek operation failed
                             * 80 = x x Drive timed out, assumed not ready
                             * AA = x   Drive not ready
                             * BB = x   Undefined error
                             * CC = x   Write fault
                             * E0 = x   Status error
                             * FF = x   Sense operation failed ~
                             *
                             * ~ = Documented for PS/2 & Extended BIOS only
                             */
    byte    n_fixed_disks;
    byte    fdd_control;
    byte    fd_controller_port;
    byte    lpt1_timeout;   /* printer 1 timeout value */
    byte    lpt2_timeout;
    byte    lpt3_timeout;
    byte    lpt4_timeout;
    byte    com1_timeout;   /* COM1 timeout value */
    byte    com2_timeout;
    byte    com3_timeout;
    byte    com4_timeout;
    word    kbd_buff_start; /* keyboard buffer start offset pointer */
    word    kbd_buff_end;   /* keyboard buffer end offset pointer */
    byte    v_rows;         /* video rows (zero relative) */
    word    char_height;    /* character height (bytes/char) */
    byte    video_ctrl_state_1;
    byte    video_ctrl_state_2;
    word    bios_40_89;     /* reserved */
    struct  /* (AT, XT after 1/10/85, PS/2) */
        {
        unsigned reserved    :4;
        unsigned step_rate   :2; /* last floppy step rate */
        unsigned data_rate   :2; /* last floppy data rate */
        } media_control;
    byte    fixed_disk_controller_status;   /* (AT, XT after 1/10/85, PS/2) */
    byte    fixed_disk_cntrlr_error_status;   /* (AT, XT after 1/10/85, PS/2) */
    byte    fixed_disk_interrupt_control;   /* (AT, XT after 1/10/85, PS/2) */
    byte    bios_40_8F;     /* reserved */
    struct MS   drive0_ms;
    struct MS   drive1_ms;
    word    bios_40_92;     /* reserved */
    byte    drive0_cyl;
    byte    drive1_cyl;
    struct
        {
        unsigned E1              :1; /* last code was E1 hidden code */
        unsigned E0              :1; /* last code was E0 hidden code */
        unsigned r_ctrl          :1; /* right ctrl key down */
        unsigned r_alt           :1; /* right alt key down */
        unsigned kbd_101         :1; /* 101/102 key keyboard installed */
        unsigned force_numlock   :1; /* force num lock if read ID & KBX */
        unsigned ID_byte_1       :1; /* last char was first ID char */
        unsigned read_in_progress:1; /* read ID in progress */
        } kbd_mode;
    struct
        {
		unsigned scroll_lock     :1;
        unsigned num_lock        :1;
        unsigned caps_lock       :1;
        unsigned reserved        :1; /* reserved, must be 0 */
        unsigned ack_received    :1;
        unsigned resend_receive  :1; /* resend receive flag */
        unsigned mode_update     :1; /* mode indicator update */
        unsigned xmit_error      :1; /* kbd transmit error flag */
        } kbd_LED;
    word    user_wait_offset; /* offset to user wait complete flag */
    word    user_wait_segment;/* segemnt to user wait complete flag */
    word    user_wait_count_low;    /* user wait count (low word) */
    word    user_wait_count_high;   /* user wait count (high word) */
    struct
        {
		unsigned int15_func86    :1; /* int 0x15 function 0x86 (Wait) has occurred */
		unsigned reserved        :6;
		unsigned elapsed         :1; /* wait time elapsed and post */
        } wait_active_flag;
	byte    bios_40_A1[7];      /* reserved */
    dblwd   video_parm_table;   /* ptr to video param table (EGA & PS/2 only) */
    dblwd   dynamic_save_area;  /* ptr to DSA (EGA & PS/2 only) */
    dblwd   alpha_aux_gen;      /* alpha mode aux char gen ptr (EGA & PS/2) */
    dblwd   graph_aux_gen;      /* graphics mode aux char gen ptr (EGA & PS/2) */
    dblwd   save_ptr_2;         /* secondary save pointer (PS/2 except M/30) */
	byte    bios_40_BC[4];      /* reserved */
	byte    bios_40_C0[64];     /* reserved */
    byte    prtscr_status;      /* print screen status byte */
	};

main ()
{
	struct BIOS_INFO *base;

	base = MK_FP (0x0040, 0x0000);
}
