        name    GETCPU  
        page    55,132
        title   "GETCPU - Determines which Intel CPU is installed"
;----------------------------------------------------------------------
;
; This program determines which Intel CPU is being used in the machine,
; whether it is an 8088/86, 80188/186, 80286 or 80386.  It uses documented 
; and supported differences in flag register bit configurations to determine 
; whether the CPU is an 80286 or 80386, and differences in shifting using CL 
; to determine if it is an 8088/86 or 80188/186.  It is intended to be used
; as an external routine from Turbo C, and returns an integer result in the
; form of the last three digits being equal to the processor type, as shown
; in the table below:
;
;       Processor Type                 Return Value
;       --------------                 ------------
;           80386                           386
;           80286                           286
;           80186/188                       186
;            8086                            86
;            8088                            88
;
;           80387                           387 << 8
;           80287                           287 << 8
;            8087                            87 << 8
;
; This code is designed for the TINY/SMALL memory model.
;
; To declare this routine in Turbo C, include the following line:
;
;       int GETCPU ();
;
; To assemble:
;
;       TASM GETCPU
;
;--------------------------------------------------------------------
;       EQUates
;--------------------------------------------------------------------
;
_80386  EQU     386
_80286  EQU     286
_80186  EQU     186
_80188  EQU     188
_8088   EQU      88
_8086   EQU      86
;
;
;--------------------------------------------------------------------
;       CODE SEGMENT
;--------------------------------------------------------------------
;
_TEXT   segment byte public 'CODE'     ; Required by TC for small model
        assume  cs:_TEXT               ; Ditto
;
;--------------------------------------------------------------------
;       Actual ID routine begins here
;--------------------------------------------------------------------
;
        public  _GETCPU                ; Define Turbo C entry point
_GETCPU proc    far                    ; Actual TC entry point
        pushf                          ; Save flags
        xor     ax,ax                  ; zero AX
        push    ax                     ; push zero onto stack
        popf                           ; put zeros into flag register
        pushf                          ; attempt to set bits 12-15 of flags
        pop     ax                     ; recover flag bits
        and     ax,08000h              ; if bits 12-15 are zero
        cmp     ax,08000h              ; then this is 8086/88 or 186/188
        jz      _8x_18x
;
;--------------------------------------------------------------------
;       Either 80286 or 80386, let's see which one
;--------------------------------------------------------------------
;
        mov     ax,07000h              ; try to set bits 12-14 to ones
        push    ax                     ; to stack
        popf                           ; to flags
        pushf                          ; back to stack
        pop     ax                     ; finally to AX
        and     ax,07000h              ; if bits 12-14 are zero
        jz      _286                   ; then it is an 80286
;
;-------------------------------------------------------------------
;       It is an 80386.
;-------------------------------------------------------------------
;
        mov     ax,_80386              ; say is 80386
        jmp     DONE                   ; quit
;
;--------------------------------------------------------------------
;       It is an 80286
;--------------------------------------------------------------------
;
_286:
        mov     ax,_80286              ; say it is 80286
        jmp     DONE                   ; quit
;
;--------------------------------------------------------------------
;       Determine if 8088/86 or 80188/186
;--------------------------------------------------------------------
;
_8x_18x:
        mov     ax,0FFFFh              ; set AX to all ones
        mov     cl,33                  ; try to shift 33 bit positions
        shl     ax,cl                  ; an 808x will shift 33 times
                                       ; an 8018x will shift only 1 time
        jnz     _18x                   ; shifting 33 times will leave all 
                                       ; zeros. If not zero, must be 18x.
        mov     ax,_8088               ; say it is an 8088
        jmp     DONE                   ; quit
;
;--------------------------------------------------------------------
;       It is an 80186/188
;--------------------------------------------------------------------
;
_18x:   
        mov     ax,_80186              ; say it is an 80186
;
;--------------------------------------------------------------------
;       Return to user with processor type in AL
;--------------------------------------------------------------------
;
DONE:
        popf                           ; restore TC's flags
        retf                           ; return to TC
_GETCPU endp
_TEXT   ends
        end     _GETCPU

