#include <dos.h>

typedef unsigned char   byte;
typedef unsigned int    word;
typedef unsigned short  pointer;
typedef unsigned long   dblwd;

#define BIOS_SEG        0x0040
#define BIOS_OFF        0x0000

#define BIOS_DATE       MK_FP(0xF000,0xFFF5)
#define BIOS_SIGN       MK_FP(0xF000,0xE005)

#define BIOS_SYSID_SEG  0xF000
#define BIOS_SYSID_OFF  0xFFFE

#define MAX_SIGNATURE   75
#define MAX_SYSID       25

struct  BIOS
    {
    char        date[9];                /* BIOS date stamp */
    char        sign[MAX_SIGNATURE];    /* BIOS OEM Signature */
    char        hard_disk_sign[MAX_SIGNATURE];
    byte        AT;                     /* TRUE if an AT; else FALSE */
    byte        sysid_model;            /* BIOS system ID byte */
    byte        sysid_submodel;
    byte        sysid_revision;
    char        sysid_char[MAX_SYSID];  /* BIOS system ID string */
    char        bus[20];                /* Bus architecture */
    char        com1[5];                /* Address for COM1: */
    char        com2[5];                /* Address for COM2: */
    char        com3[5];                /* Address for COM3: */
    char        com4[5];                /* Address for COM4: */
    char        lpt1[5];                /* Address for LPT1: */
    char        lpt2[5];                /* Address for LPT2: */
    char        lpt3[5];                /* Address for LPT3: */
    char        lpt4[5];                /* Address for LPT4: */
    byte        n_floppies;             /* Number of floppy drives */
    byte        n_fixed_disks;          /* Number of hard drives */
    word        drive0_data_rate;
    word        drive1_data_rate;
    char        int09[4];
    char        chip2[4];
    char        ext_wait[4];
    };

struct VIDEO
    {
    char    adapter[20];
    char    sign[MAX_SIGNATURE];
    };
/*
 * Media Parameter Table.
 *
 * Applies to AT after 11/15/85, XT after 01/10/86, XT286 and PS/2
 *
 * Int 0x13 function 0x08 returns pointer to MPT.
*/

struct MPT
    {
    byte    specify_1;      /* First specify byte */
    byte    specify_2;      /* Second specify byte
                             * 1 = 360K floppy
                             * 2 = 1.2M floppy
                             * 3 = 720K floppy
                             * 4 = 1.44M floppy
                            */
    byte    ticks_off;      /* timer ticks to wait until motor off */
    byte    bytes_sector;   /* number of bytes per sector
                             * 0 = 128 bps
                             * 1 = 256 bps
                             * 2 = 512 bps
                             * 3 = 1024 bps
                            */
    byte    sectors_track;  /* number of sectors per track */
    byte    gap_len;        /* gap length, in bytes */
    byte    data_len;       /* data length, in bytes */
    byte    format_gap;     /* gap length for format */
    byte    format_fill;    /* fill byte for format */
    byte    settle_time;    /* head settle time, in milliseconds */
    byte    start_time;     /* motor start time, in 1/8 seconds */
    };

struct  dev_hdr
   {
   struct  dev_hdr     *dh_next;   /* Next device pointer */
   struct  dev_attr
       {
       unsigned std_in      : 1;    /* Standard input device */
       unsigned std_out     : 1;    /* Standard output device */
       unsigned nul_dev     : 1;    /* NUL device */
       unsigned clock       : 1;    /* CLOCK$ device */
       unsigned int_29h     : 1;    /* driver services Int 29h */
       unsigned rsrvd       : 1;    /* Reserved */
       unsigned gen_ioctl   : 1;    /* DOS 3.2+ Supports generic IOCTL */
       unsigned rsrvd1      : 4;    /* Reserved */
       unsigned removable   : 1;    /* Supports OPEN/CLOSE/Removable media */
       unsigned rsrvd2      : 1;    /* Reserved */
       unsigned out_busy    : 1;    /* Char: Does not support output till busy */
                              /* Block: IBM Block Format */
       unsigned ioctl       : 1;    /* IOCTL supported */
       unsigned char_dev    : 1;    /* 1 = Character device, 0 = Block device */
       } attribs;
   pointer      dh_strat;  /* Driver startegy routine */
   pointer      dh_inter;  /* Driver interrupt routine */
   char         dh_name[8];     /* Device name */
};

struct DPB              /* Drive Parameter Block */
{
   byte            drive;          /* 00h drive number (zero relative) */
   byte            unit;           /* 01h device driver unit number */
   int             bps;            /* 02h bytes per sector */
   byte            spc;            /* 04h sectors per cluster (zero relative) */
   byte            shift;          /* 05h shift factor (power of two) */
   int             rsrvd;          /* 06h # reserved boot sectors */
   byte            fats;           /* 08h # of FAT copies */
   word            rootdir;        /* 09h # root directory entries */
   word            sector1;        /* 0Bh first data sector number */
   word            sectorn;        /* 0Dh highest cluster number + 1 */
   byte            spf;            /* 0Fh sectors per FAT */
   word            root1;          /* 10h root directory starting sector # */
   struct  devhdr  *devadr;        /* 12h device driver address */
   byte            media;          /* 16h media descriptor byte */
   byte            dpbflag;        /* 17h DPB access flag (0xFF=need to rebuild) */
   struct  DPB     *next;          /* 18h pointer to next DPB */
   word            lclust;         /* 1Ch last cluster number allocated */
   word            unk0;           /* 1Eh purpose unknown - normally 0xFFFF */
};

struct DCB              /* Device Control Block */
{
   int             nusers;         /* 00h # users for this DCB */
   int             mode;           /* 02h 0,1,2 per OPEN */
   byte            datrb;          /* 04h disk attribute byte */
   byte            dvatr;          /* 05h device attribute byte (hi) */
   byte            atrb2;          /* 06h 2nd device attribute */
   struct  devhdr         *devadr; /* 07h driver address */
   word            frstc;          /* 0Bh first cluster # */
   word            modtm;          /* 0Dh file time word */
   word            moddt;          /* 0Fh file date word */
   long            totsiz;         /* 11h total file size */
   long            curpos;         /* 15h current byte position */
   word            clsctr;         /* 19h total cluster counter */
   word            curcls;         /* 1Bh current cluster number */
   word            dirsec;         /* 1Dh directory sector */
   byte            dirndx;         /* 1Fh directory entry index (0..n) */
   char            name[8];        /* 20h device/file name */
   char            ext[3];         /* 28h file extension */
   word            unk0;           /* 2Bh unknown */
   word            unk1;           /* 2Dh unknown */
   word            unk2;           /* 2Fh unknown */
   word            owner;          /* 31h PSP of owner process */
   word            unk3;           /* 33h unknown */
   word            unk4;           /* 35h unknown */
   word            unk5;           /* 37h unknown */
   word            unk6;           /* 39h unknown */
};

struct LDT              /* Logical Drive Table */
{
   char        drive[2];       /* 00h drive letter + ':' */
   char        path[65];       /* 02h current path for this drive */
   struct                      /* 43h current drive status */
       {
       unsigned LDT_UNUSED         :11;
       unsigned LDT_status_unmapped:1;
       unsigned LDT_status_subst   :1;  /* Device is substituted */
       unsigned LDT_status_unknown1:1;
       unsigned LDT_status_ready   :1;  /* Device is ready */
       unsigned LDT_status_unknown :1;
       } status;
   struct  DPB *dpb;           /* 45h pointer to DPB for this drive */
   word        cluster;        /* 49h 1st cluster of current directory */
                                      /* 0xFFFF if never accessed */
                                      /* 0x0000 if at root directory */
   word        unk0;           /* 4Bh unknown usage */
   word        unk1;           /* 4Dh unknown usage */
   word        SUBSTlen;       /* 4Fh SUBST path length */
};

struct LINK             /* Chain link for DCB, FCB chains */
{
   void        *LINK;          /* 00h pointer to next link or 0xFFFF */
   int         number;         /* 04h number of blocks here */
};

struct CVT              /* Configuration Variable Table */
{
   void            *curuff;  /* -08h current buffer pointer */
   word            buffofs;  /* -04h offset in current buffer */
   word            mcb_seg;  /* -02h segment of first MCB */
   struct  DPB     *dpb;     /*  00h pointer to first Drive Parameter Block */
   struct  DCB     *dcb;     /*  04h pointer to first Device Control Block */
   void            *clock;   /*  08h pointer to CLOCK$ device driver */
   void            *con;     /*  0Ch pointer to CON device driver */
   word            maxbps;   /*  10h max bytes per sector on any block device */
   struct LINK     *s_dbc;   /*  12h pointer to start of disk buffer chain */
   struct  LDT     *ldt;     /*  16h pointer to logical drive table */
   struct  fcb     *s_fcb;   /*  1Ah pointer to start of FCB chain */
   word            nfcb;     /*  1Eh # FCB's to keep when swapping */
   byte            nblkdev;  /*  20h number of block devices */
   byte            ndrives;  /*  21h number of logical drives */
   struct  dev_hdr *Null;    /* 22h NUL device driver */
};

struct MCB              /* Memory allocation block header format */
{
   char    flag;           /* 00h must be 'M' or 'Z' */
   word    psp;            /* 01h PSP of owner or 0x0000 */
   word    size;           /* 03h number of paragraphs */
   char    junk[3];        /* 05h unused */
   char    owner[8];       /* 08h name of owner */
};

struct  SDT_feat        /* feature info */
{
   unsigned reserved : 1;    /* reserved bit */
   unsigned isa_mca  : 1;    /* 0 = ISA, 1 = MCA */
   unsigned ebda     : 1;    /* 1 = Extended BIOS data area allocated */
   unsigned ext_wait : 1;    /* 1 = Wait for external event supported */
   unsigned kbd_int  : 1;    /* 1 = kbd intercept called by int 0x09 */
   unsigned rt_clock : 1;    /* 1 = real time clock is present */
   unsigned chip_2   : 1;    /* 1 = 2nd interrupt chip present */
   unsigned dma_3    : 1;    /* 1 = DMA channel 3 used by hard disk BIOS */
};

/*
 * System Descriptor Table
 *
 * Address returned by int 0x15 function 0xC0
*/
struct  SDT
{
   word    count;          /* length of sdt */
   byte    model;          /* model identification */
   byte    submodel;       /* submodel identification */
   byte    rev_level;      /* BIOS revision level (00 = 1st release) */
   struct  SDT_feat     features;       /* features byte */
   byte    reserved[4];    /* reserved bytes */
};

struct EQUIP
    {
    unsigned disk   : 1;    /* Disk Drive Installed */
    unsigned math   : 1;    /* Math co-processor installed */
    unsigned ram    : 2;    /* Motherboard RAM */
    unsigned video  : 2;    /* init video mode */
    unsigned drives : 2;    /* number of floppies (zero relative) */
    unsigned spare  : 1;    /* unused */
    unsigned serial : 3;    /* number of serial ports */
    unsigned game   : 1;    /* game port installed */
    unsigned filler : 1;    /* not used */
    unsigned prntrs : 2;    /* number of printer ports */
    };

struct LIM_EMS
    {
    int     version;
    int     free;
    int     total;
    };

/*
 * Media State byte definition as used by BIOS Drive 0 (and 1) media state
 * bytes at addresses 40:90 & 40:91.
*/
struct  MS
    {
    unsigned media_state         :3; /* not currently documented */
    unsigned reserved            :1;
    unsigned media_established   :1;
    unsigned double_step_required:1;
    unsigned data_rate           :2; /* 00 = 500KB/s, 01 = 300KB/s, 10 = 250KB/s */
    };

struct BIOS_INFO
    {
    word    com1;
    word    com2;
    word    com3;
    word    com4;
    word    lpt1;
    word    lpt2;
    word    lpt3;
    word    lpt4;
    struct
        {
        unsigned is_floppies :1;
        unsigned is_math     :1;
        unsigned is_mouse    :1;    /* PS/2 only */
        unsigned reserved    :1;    /* reserved  */
        unsigned video_mode  :2;    /* 01=CO40, 10=CO80, 11=BW80 */
        unsigned n_floppies  :2;
        }   hardware_1;
    struct
        {
        unsigned reserved    :1;    /* reserved */
        unsigned n_rs232     :3;
        unsigned is_joystick :1;
        unsigned is_modem    :1;    /* Convertible only */
        unsigned n_lpt       :2;
        }   hardware_2;
    byte    POST_status;            /* Convertible only */
    word    mem_size;               /* in K bytes */
    word    bios_40_15;             /* reserved */
    struct
        {
        unsigned r_shift     :1;
        unsigned l_shift     :1;
        unsigned ctrl        :1;
        unsigned alt         :1;
        unsigned scroll_lock :1; /* scroll lock mode */
        unsigned num_lock    :1; /* num lock mode */
        unsigned caps_lock   :1; /* caps lock mode */
        unsigned insert      :1; /* insert mode */
        }   kbd_1;
    struct
        {
        unsigned l_ctrl      :1; /* Left ctrl key is down */
        unsigned l_alt       :1;
        unsigned SysReq      :1;
        unsigned pause       :1; /* pause mode active */
        unsigned scroll_lock :1; /* scroll lock key down */
        unsigned num_lock    :1; /* num lock key down */
        unsigned caps_lock   :1; /* caps lock key down */
        unsigned insert      :1; /* insert key down */
        }   kbd_2;
    byte    alt_keypad;     /* alternate keypad entry */
    word    kbd_head;       /* keyboard buffer head */
    word    kbd_tail;       /* keyboard buffer tail */
	byte    kbd_buff[32];   /* keyboard buffer */
    struct
        {
        unsigned drive0      :1;
        unsigned drive1      :1;
        unsigned drive2      :1;
        unsigned drive3      :1;
        unsigned reserved    :3;
        unsigned intrupt     :1;
        } recallibrate_status;
    struct
        {
        unsigned drive0      :1;
        unsigned drive1      :1;
        unsigned drive2      :1;
        unsigned drive3      :1;
        unsigned selected    :2;
        unsigned reserved    :1;
        unsigned write_read  :1;
        } motor_status;
    byte    motor_off_count;
    struct
        {
        unsigned bad_drive   :1; /* invalid drive parameter */
        unsigned bad_address :1; /* address mark not found */
        unsigned sector_nf   :1; /* sector not found */
        unsigned DMA_overrun :1; /* DMA overrun on operation */
        unsigned CRC_error   :1; /* CRC error on diskette read */
        unsigned general_err :1; /* General controller failure */
        unsigned seek_failed :1; /* Seek operation failed */
        unsigned not_ready   :1; /* Drive not ready */
        } floppy_prev_op_status;
	byte    floppy_controller_status[7];
    byte    display_mode;       /* current video mode */
    word    n_cols;             /* number of video columns */
    word    regen_len;
    word    regen_addr;
    word    cursor_pos_page_1;
    word    cursor_pos_page_2;
    word    cursor_pos_page_3;
    word    cursor_pos_page_4;
    word    cursor_pos_page_5;
    word    cursor_pos_page_6;
    word    cursor_pos_page_7;
    word    cursor_pos_page_8;
    word    cursor_type;
    byte    current_page;
    word    video_controller_base_addr;
    byte    curr_3x8_register;
    byte    curr_3x9_register;
    byte    *reset_code;    /* PS/2 only (except Model 30) */
    byte    bios_40_6B;     /* reserved */
    dblwd   timer_count;    /* timer counter */
    byte    timer_oflow_flag;
    byte    break_key_state;
    word    reset_flag;     /* 0x1234 = bypass memory test
                             * 0x4321 = preserve memory (PS/2)
                             * 0x5678 = system suspended (convertible)
                             * 0x9ABC = manufacturer test (convertible)
                             * 0xABCD = system post loop (convertible)
                            */
    byte    prev_op_status; /*      H/F (Hard/Floppy/Both)
                             * 00 = x x No error
                             * 01 = x x Bad Command
                             * 02 = x x Address mark not found
                             * 03 =   x Write protect error
                             * 04 = x x Sector not found
                             * 05 = x   Reset failed
                             * 06 =   x Diskette was removed
                             * 07 = x   Bad parameter table
                             * 08 =   x DMA overrun
                             * 09 = x x Attempted to cross 64k boundary on DMA
                             * 0A = x   Bad sector flag
                             * 0B = x   Bad cylinder detected ~
                             * 0C =   x Requested Media type not found ~
                             * 0D = x   Invalid # sectors in format ~
                             * 0E = x   Control data address mark detected ~
                             * 0F = x   DMA arbitration level out of range ~
                             * 10 = x x CRC or ECC error on disk read
                             * 11 = x   ECC corrected data error
                             * 20 = x x Controller failed
                             * 40 = x x Seek operation failed
                             * 80 = x x Drive timed out, assumed not ready
                             * AA = x   Drive not ready
                             * BB = x   Undefined error
                             * CC = x   Write fault
                             * E0 = x   Status error
                             * FF = x   Sense operation failed ~
                             *
                             * ~ = Documented for PS/2 & Extended BIOS only
                            */
    byte    n_fixed_disks;  /* Number of hard disks installed */
    byte    fdd_control;
    byte    fd_controller_port;
    byte    lpt1_timeout;   /* printer 1 timeout value */
    byte    lpt2_timeout;
    byte    lpt3_timeout;
    byte    lpt4_timeout;
    byte    com1_timeout;   /* COM1 timeout value */
    byte    com2_timeout;
    byte    com3_timeout;
    byte    com4_timeout;
    word    kbd_buff_start; /* keyboard buffer start offset pointer */
    word    kbd_buff_end;   /* keyboard buffer end offset pointer */
    byte    v_rows;         /* video rows (zero relative) */
    word    char_height;    /* character height (bytes/char) */
    byte    video_ctrl_state_1;
    byte    video_ctrl_state_2;
    word    bios_40_89;     /* reserved */
    struct  /* (AT, XT after 1/10/85, PS/2) */
        {
        unsigned reserved    :4;
        unsigned step_rate   :2; /* last floppy step rate */
        unsigned data_rate   :2; /* last floppy data rate */
        } media_control;
    byte    fixed_disk_controller_status;   /* (AT, XT after 1/10/85, PS/2) */
    byte    fixed_disk_cntrlr_error_status;   /* (AT, XT after 1/10/85, PS/2) */
    byte    fixed_disk_interrupt_control;   /* (AT, XT after 1/10/85, PS/2) */
    byte    bios_40_8F;     /* reserved */
    struct MS   drive0_ms;
    struct MS   drive1_ms;
    word    bios_40_92;     /* reserved */
    byte    drive0_cyl;
    byte    drive1_cyl;
    struct
        {
        unsigned E1              :1; /* last code was E1 hidden code */
        unsigned E0              :1; /* last code was E0 hidden code */
        unsigned r_ctrl          :1; /* right ctrl key down */
        unsigned r_alt           :1; /* right alt key down */
        unsigned kbd_101         :1; /* 101/102 key keyboard installed */
        unsigned force_numlock   :1; /* force num lock if read ID & KBX */
        unsigned ID_byte_1       :1; /* last char was first ID char */
        unsigned read_in_progress:1; /* read ID in progress */
        } kbd_mode;
    struct
        {
        unsigned scroll_lock     :1;
        unsigned num_lock        :1;
        unsigned caps_lock       :1;
        unsigned reserved        :1; /* reserved, must be 0 */
        unsigned ack_received    :1;
        unsigned resend_receive  :1; /* resend receive flag */
        unsigned mode_update     :1; /* mode indicator update */
        unsigned xmit_error      :1; /* kbd transmit error flag */
        } kbd_LED;
    word    user_wait_offset; /* offset to user wait complete flag */
    word    user_wait_segment;/* segemnt to user wait complete flag */
    word    user_wait_count_low;    /* user wait count (low word) */
    word    user_wait_count_high;   /* user wait count (high word) */
    struct
        {
        unsigned int15_func86    :1; /* int 0x15 function 0x86 (Wait) has occurred */
        unsigned reserved        :6;
        unsigned elapsed         :1; /* wait time elapsed and post */
        } wait_active_flag;
    byte    bios_40_A1[7];      /* reserved */
    dblwd   video_parm_table;   /* ptr to video param table (EGA & PS/2 only) */
    dblwd   dynamic_save_area;  /* ptr to DSA (EGA & PS/2 only) */
    dblwd   alpha_aux_gen;      /* alpha mode aux char gen ptr (EGA & PS/2) */
    dblwd   graph_aux_gen;      /* graphics mode aux char gen ptr (EGA & PS/2) */
    dblwd   save_ptr_2;         /* secondary save pointer (PS/2 except M/30) */
    byte    bios_40_BC[4];      /* reserved */
    byte    bios_40_C0[64];     /* reserved */
    byte    prtscr_status;      /* print screen status byte */
    };
